from __future__ import annotations

"""Perk metadata and deterministic perk runtime modules."""

from .ids import (
    PERK_BY_ID,
    PERK_DEFAULT_FLAGS,
    QUICK_LEARNER_DESCRIPTION,
    QUICK_LEARNER_NAME,
    PerkFlags,
    PerkId,
    PerkMeta,
    perk_display_description,
    perk_display_name,
    perk_label,
)

__all__ = [
    "PERK_BY_ID",
    "PERK_DEFAULT_FLAGS",
    "QUICK_LEARNER_DESCRIPTION",
    "QUICK_LEARNER_NAME",
    "PerkFlags",
    "PerkId",
    "PerkMeta",
    "perk_display_description",
    "perk_display_name",
    "perk_label",
]
